// application to extract frequently detected sets of instances
// 26 july 2009 :  
//// no internal PWM-comparison :
// ->no ScoreRankingRun.h/cpp->(no -s,-t,-x) => Cluster.h/cpp
//// no InstanceIO.h/cpp => implement in FastaIO.h/cpp
//// non-active -e,-r,-a
//// minimize functionality in matlab :
// no iMotifClustering.m  => mClustInst.m, motifreconstruction in c++ 
//// 27 july 2009 : 
// implement mClustEdgesSym() & mClustEdgesAsymTrace() and -t userparameter
//// 06 aug 2009 :
// Instance start >=0 ipv >0; TFBS-weight in BuildAsymmetricalMatrix; 
// meaningless cluster filtering in .m functions (new libClust).
//// 15 sept 2009 :
// optimalization in Cluster->ReconstructMotif and MotifClusterRun->PrintClusters()
//// 12 dec 2009 :
// major optimalization in LoadInstances() !!
// implement -j (case -t<asym>) : filter on min#inputsets
// 09/04/2010 : -p default to 0.10
// 09/11/2010 : visualize membership for asym/sym mode
// 07/03/2011 : implement asym mode as default clustering
// 07/04/2011 : minor revisions in output formatting
// 18/05/2011 : bug fixed in loading empty instances lines
// 27/06/2011 : allow '>' prefix in inputfile (default prefix = #id)
//// MAJOR REVISION 22 sept 2011
// remove unused code such as a:(BiasH), e:(Extract(motif/module)), r:(runPositional)
// and also parameter -t (only asymmetrical clustering now), and also -p (==-j)
// => check versions 5.1.x if recovery requested
// (update the naming of program parameters)
// memberships : make it externally accessible as percentual setpoint
// weights in inputfile : apply to matrix T
// parameter -j : check comparison (int)
// adjust some output formatting
// 14/03/2012 : adjust output screen messaging 
// cerrstr -> use ostringstream to file and limit the progress reporting messages

#include "libClust.h"
#include "MotifClusterRun.h"
#include <getopt.h>

#define OPTIONS "c:d:f:hi:j:m:o:p:O:v?"
// function prototypes
void instructions();
void version();
void cleanup();

// define some global variables to store input
string *instanceFile = NULL,
       *outputFile = NULL,
       *matrixFile = NULL,
       *debugFile = NULL;

int main(int argc, char *argv[])
{
  // process user-input
  char c;
  // defaults
  double MinNodeFreq(0.1); // to prefilter matrix T (make it more sparse)
  double MemberCut(-1); // to postfilter an extracted cluster (%)
  double MinCS(0.5); // minimal consensus score (-c)
  double MinSeqInst(0.5); // to postfilter on sequence-coverage
  double MinDataFreq(0.2); // and to postfilter (in absolute value)
	
//14 march 2012
  ostringstream cerrstr; // to output all error messages (instead of cerr)
  bool wronginput = false;

  if (optopt == 0)
  {
    instructions();
    exit(-1);
  }
  while ((c = getopt(argc, argv, OPTIONS)) != -1)
  {
    switch (c)
    {
    case 'f':
      instanceFile = new string(optarg);
      break;
    case 'O':
      matrixFile = new string(optarg);
      break;
    case 'o':
      outputFile = new string(optarg);
      break;
    case 'd':
      debugFile = new string(optarg);
      break;
    case 'c':
      MinCS = (double) atof(optarg);
      if ( MinCS < 0 || MinCS >= 2 )
      { cerr << "ERROR: -c input minimal consensus score should between 0 and 2.\n";
        cerrstr << "ERROR: -c input minimal consensus score should between 0 and 2.\n";
        wronginput = true;
        //cleanup(); exit(-1);
      }
      break;
    case 'i':
      MinSeqInst = (double) atof(optarg);
      if ( MinSeqInst <= 0 || MinSeqInst > 1)
      { cerr << "--ERROR: -i input should be between 0 and 1." << endl;
        cerrstr << "--ERROR: -i input should be between 0 and 1." << endl;
        wronginput = true;
        //cleanup(); exit(-1);
      }
      break;
    case 'j':
      MinDataFreq = (double) atof(optarg);
      if ( MinDataFreq  <= 0 || MinDataFreq  > 1)
      { cerr << "--ERROR: -j input should be between 0 and 1." << endl;
        cerrstr << "--ERROR: -j input should be between 0 and 1." << endl;
        wronginput = true;
        //cleanup(); exit(-1);
      }
      break;
    case 'm':
      MemberCut = (double) atof(optarg);
      if ( (MemberCut < 0 && MemberCut != -1) || MemberCut >= 1)
      { cerr << "--ERROR: -m input (percentual) membership threshold should be between 0 and 1 "
             << "or -1 for optimization-based cutoff" << endl; 
        cerrstr << "--ERROR: -m input (percentual) membership threshold should be between 0 and 1 "
             << "or -1 for optimization-based cutoff" << endl;
        wronginput = true;
        //cleanup(); exit(-1);
      }
      break;
    case 'p':
      MinNodeFreq = (double) atof(optarg);
      if ( MinNodeFreq  < 0 || MinNodeFreq  >= 1)
      { cerr << "--ERROR: -p input should be between 0 and 1." << endl;
        cerrstr << "--ERROR: -p input should be between 0 and 1." << endl;
        wronginput = true;
        //cleanup(); exit(-1);
      }
      break;
    case 'v':
      version(); exit(-1);
    case '?':
    case 'h':
      instructions(); exit(-1);
    default:
      cerr << "FuzzyClustering: Error in input getopt() function" << endl; 
      cerrstr << "FuzzyClustering: Error in input getopt() function" << endl; 
      wronginput = true;
      //cleanup(); exit(-1);
    }
  }

  // check wrong input if so report/exit
  GFFWriter * outputText = NULL;
  if (wronginput)
  { 
    // report to file if file is available
    if (outputFile != NULL) 
    { cerrstr << "#Check our MotifSuite webpage for correct input format. Or contact us, we will be happy to help." << endl;
      outputText = new GFFWriter(outputFile); 
      outputText->AddComment(cerrstr.str());
      cerrstr.flush(); // flush  
      if (outputText != NULL) delete outputText; outputText = NULL;
    }
    // exit
    cleanup(); exit(-1);  
  }

  // check if required input is available
  if (instanceFile == NULL)
  { 
    cerr << "--ERROR: -f input file is not provided." << endl; 
    cerrstr << "--ERROR: -f input file is not provided." << endl; 
    instructions(); 
    // report to file if file is available
    if (outputFile != NULL) 
    { cerrstr << "#Check our MotifSuite webpage for correct input format. Or contact us, we will be happy to help." << endl;
      outputText = new GFFWriter(outputFile); 
      outputText->AddComment(cerrstr.str());
      cerrstr.flush(); // flush  
      if (outputText != NULL) delete outputText; outputText = NULL;
    }
    // exit
    cleanup(); exit(-1); 
  }

  // Load MCR and library 
  cerr << "#Initialize MCR and Matlab library." << endl;
  if (!mclInitializeApplication(NULL,0) ) 
  { cerr << "--Error-- could not initialize the MCR application properly.\n";
    cerrstr << "--Error-- could not initialize the MCR application properly.\n";
    // report to file if file is available
    if (outputFile != NULL) 
    { cerrstr << "#Contact us, we will be happy to help." << endl;
      outputText = new GFFWriter(outputFile); 
      outputText->AddComment(cerrstr.str());
      cerrstr.flush(); // flush  
      if (outputText != NULL) delete outputText; outputText = NULL;
    }
    // exit
    cleanup(); exit(-1);  
  }
  if( !libClustInitialize() )
  { cerr << "--Error-- could not initialize the Matlab library properly.\n ";
    cerrstr << "--Error-- could not initialize the Matlab library properly.\n ";
    // report to file if file is available
    if (outputFile != NULL) 
    { cerrstr << "#Contact us, we will be happy to help." << endl;
      outputText = new GFFWriter(outputFile); 
      outputText->AddComment(cerrstr.str());
      cerrstr.flush(); // flush 
      if (outputText != NULL) delete outputText; outputText = NULL; 
    }
    // exit
    cleanup(); exit(-1); 
  }

  // open output for writing
  // text
  if (outputText == NULL)  
  { if (outputFile != NULL) outputText = new GFFWriter(outputFile);
    else outputText = new GFFWriter(); // to screen if no file is supplied
  }
  GFFWriter * debugText = NULL; 
  if (debugFile != NULL) debugText = new GFFWriter(debugFile);
  // matrices
  PWMIO * outputMatrix = NULL; 
  if (matrixFile != NULL) outputMatrix = new PWMIO(matrixFile, WRITE);

  // send user-input to outputFile
  // 06 april 2011 : do not send command to -o file 
  ostringstream usr; 
  //07 april 2011
  usr << "#The first section of this file allows you to evaluate " ;
  usr << "(#DatasetDimension:) the amount of noise in your input, " ;
  usr << "(#List of merged non-spurious instances:) variability of the instances width in your input, " ;
  usr << "(#MotifClusterRun:) post-processing of the extracted clusters." << endl;
  usr << "#The section #PrintClusters describes the ensemble motif(s) extracted by the FuzzyClustering application." << endl;
  usr << "#Please read more on FuzzyClustering Guidelines webpage to optimally interpretate below results." << endl;

  // add to file/screen
  outputText->AddComment(usr.str());
  usr.flush(); // flush  
  
  MotifClusterRun * pMainRunner = NULL;
  try 
  {
    // create an object 
    pMainRunner = new MotifClusterRun();
    pMainRunner->LinkoutputText(outputText, debugText);

    cerr << "#Loading instances from " << *instanceFile << endl;
    pMainRunner->LoadInstances(instanceFile); 
    // check
    pMainRunner->DatasetDimension(MinNodeFreq, MinSeqInst, MinDataFreq);

    cerr << "#Start fuzzy clustering procedure" << endl;
    try
    { 
      // cluster high frequency nodes
      if (pMainRunner->ClusterXnodesYdata(MemberCut, MinCS))
      {
        // print results to output (-o)
        pMainRunner->PrintClusters();
        // and to matrixfile if applicable (-O)
        if (outputMatrix != NULL)
        {
          pMainRunner->PrintMotifs(outputMatrix);
        }
        cerr << "#End of fuzzy clustering procedure." << endl;
      }
      else
      {
        cerr << "#Unsuccessful end of fuzzy clustering procedure." << endl;
        cerrstr << "#Unsuccessful end of fuzzy clustering procedure. " << 
            "Retry with revised input/parameters or contact us." <<  endl;
        outputText->AddComment(cerrstr.str());
        cerrstr.flush(); // flush  
      }
    }
    catch (const mwException& e)
    { cerr << "--Error--catch : " << e.what() << endl; 
      cerrstr << "--Error--catch : " << e.what() << endl;
      outputText->AddComment(cerrstr.str());
      cerrstr.flush(); // flush   
    } //return -2;
    catch (...)
    { cerr << "--Error--catch : Unexpected error thrown" << endl;
      cerrstr << "--Error--catch : Unexpected error thrown" << endl;
      outputText->AddComment(cerrstr.str());
      cerrstr.flush(); // flush    
    } //return -3;
  }
  catch (ErrorInputfiles)
  { cerr << "--Error--catch ErrorInputfiles: Ending program without results." << endl;
    cerrstr << "--Error--catch ErrorInputfiles: Ending program without results." << endl;
    outputText->AddComment(cerrstr.str());
    cerrstr.flush(); // flush  
  }
  catch (InputDimension)
  { cerr << "--Error--catch ErrorDimension: Ending program without results." << endl;
    cerrstr << "--Error--catch ErrorDimension: Ending program without results." << endl;
    outputText->AddComment(cerrstr.str());
    cerrstr.flush(); // flush  
  }


  /* cleanup variables and close outputfiles--------------------------------*/
  if (pMainRunner!= NULL) delete pMainRunner; pMainRunner = NULL;
  if (outputText != NULL) delete outputText; outputText = NULL; 
  if (debugText != NULL) delete debugText; debugText = NULL; 
  if (outputMatrix != NULL) delete outputMatrix; outputMatrix = NULL;

  cerr << "#Close MCR and Matlab library." << endl;
  libClustTerminate();
  mclTerminateApplication();
  cleanup();
  cerr << "#Exit program succesfully." << endl; 
  exit(0);
}

/*---------------------------------------------------------------------------*/
void
instructions()
{
  cout << endl;
  cout << "Usage:" << " FuzzyClustering " << endl;
  cout << endl;
  cout << " Required Arguments" << endl;
  cout << "  -f <instancefile>    List of instances, grouped by motif detection run.\n"; 
  cout << endl;
  cout << " Optional Arguments" << endl;
  cout << "  -c <value>          Minimal consensus score of an ensemble motif (default 0.5)\n";
  cout << "  -i <value>          Minimal fraction of sequences where the ensemble ";
  cout << "motif has at least one instance (default 0.5)." << endl;
  cout << "  -j <value>          Minimal fraction of input motifs that corresponds to the ensemble motif ";
  cout << "(default 0.20)." << endl;
  cout << "  -m <value>          Percentual membership cutoff (default computed by the program)\n";
  cout << "  -p <value>          Minimal detection frequency of an instance in the supplied data (default 0.10)\n";
  cout << " Output Arguments" << endl;
  cout << "  -o <outputFile>     Output file to write motifs in instances format (default stdout)." << endl;
  cout << "  -O <matrixFile>     Output file to write motifs in pwm format (default not used)." << endl;
  cout << "  -d <debugFile>      Output file for development (debug)." << endl;
  cout << endl;
  cout << "Version " << FUZZYVERSION << endl;
  cout << "Questions and Remarks: <mclaeys@qatar.net.qa>" << endl << endl;

} 
void
version()
{
  cout << endl;
  cout << "INCLUSive -- FuzzyClustering " << endl;
  cout << "  version " << FUZZYVERSION << endl;
    cout << endl;
  cout << "  Revision history : " << endl;
  cout << "  - 06/08/2009 : minor debug updates" << endl;
  cout << "  - 15/09/2009 : optimization in Cluster->ReconstructMotif" << endl;
  cout << "  - 12/12/2009 : major optimization in MotifClusterRun->LoadInstances," << endl;
  cout << "                 implement -j (case -t<asym>) : filter on min#inputsets" << endl;
  cout << "  - 09/04/2010 : -p default to 0.10" << endl;
  cout << "  - 09/11/2010 : visualize membership for asym/sym mode" << endl;
  cout << "  - 07/03/2011 : implement asym mode as default clustering" << endl;
  cout << "  - 07/04/2011 : minor revisions in output formatting" << endl;
  cout << "  - 18/05/2011 : bug fixed in loading empty instances lines" << endl;
  cout << "  - 27/06/2011 : allow '>' prefix in inputfile (default prefix = #id)" << endl;
  cout << "  - 22/09/2011 : version 5.2.0 : external access membership cutoff, ";
  cout << "weights in matrix T, -j cutoff (INT) , some output formatting" << endl;
  cout << "  - 14/03/2012 : visualize error messages to user file." << endl;
  cout << "  - 29/10/2013 : fix error throw in InstanceSet::Datadimension()--RemoveLowProbIndividuals()." << endl;
  cout << "  -end." << endl;
  cout << endl;
}
void
cleanup()
{
  if (instanceFile != NULL)
    delete instanceFile;
  instanceFile = NULL;
  if (outputFile != NULL)
    delete outputFile;
  outputFile = NULL;
  if (matrixFile != NULL)
    delete matrixFile;
  matrixFile = NULL;
  if (debugFile != NULL)
    delete debugFile;
  debugFile = NULL;

}
