#ifndef cluster_include_declared
#define cluster_include_declared

#include "fuzzutilities.h" //#include "inclusive.h"
#include "PWM.h"
#include "InstanceSet.h"

class Cluster{
 private:
  // node membership of cluster
  double * _pCluster;
  int _length1;
  int _nbrNodes; // number of non-zeros
  int _nbrNodesEx; // number of excluded nodes by edgeclust.m
  double _NodeThresh; // compute min/max node membership 
  // data membership of cluster
  double * _pClusterData;
  int _length2;
  int _nbrData; // number of non-zeros
  int _nbrDataEx; // number of excluded data by edgeclust.m
  double _DataThresh; // compute min/max node membership 

  // cluster in motifmatrix format
  PWM * _pGoodCSMotif;
  PWM * _pGoodwidthMotif;
  double * _pseudo;
  double * _pClusterShifts;
  string * _pCSdistribution;


 public:
  // constructors
  Cluster(int nbrN, double * cluster, int nbrEx1, int nbrD, double * data, int nbrEx2);

  // destructor
  ~Cluster();

  // inspectors
  int NbrNNZNodes() { return _nbrNodes;}
  int NbrNNZData() { return _nbrData;}
  double NodeThresh(){return _NodeThresh;};
  double DataThresh(){return _DataThresh;};
  double * GetClusterNodes(){return _pCluster;};
  double * GetClusterShifts(){return _pClusterShifts;};
  double * GetClusterData(){return _pClusterData;};
  PWM * GetCSMotif(){return _pGoodCSMotif;};
  PWM * GetWidthMotif(){return _pGoodwidthMotif;};
  string * GetCSdistribution(){return  _pCSdistribution;}
  int WeightedInstanceCount();
  int WeightedMotifCount();

  // motif reconstructie
  void ReconstructMotif(InstanceSet * instanceSet);
  iSetIter GetLongestHighestMemberInstance(InstanceSet * instanceSet);
  int * SearchExtensions(InstanceSet * instanceSet, int length, 
                          int ext, int shift, double CSref);
  void BuildCSdistribution(InstanceSet * instanceSet, 
                             int length, int left, int right);
  PWM * BuildPWM(InstanceSet * instanceSet, int width, int HprobShift);
  
  
};

#endif
