#ifndef instance_include_declared
#define instance_include_declared

#include "inclusive.h"
#include "SequenceObject.h"


//class SequenceObject;

class Instance{
 private:
  int _start;
  int _length;
  strand_modes _strand;
  double _score;
  string *_pId;
  Sequence *_pSite;
  string *_pSiteString;
  SequenceObject * _pParentSequence;
  string *_pSeqName;
  // FC
  string * _pMotifID; // FC
  double _weight; // FC
  vector<Instance*> * _pOverlaps; // FC
  int _pOcc; // FC
  vector<string> * _pJoinMotifIDs;// FC, to avoid repetitive GetMotifID
  vector<double> * _pJoinWeights;// FC, to avoid repetitive GetWeight
  int _iStore; // FC, to aid optimal storing of overlapping sets


 public:
  // constructors
  Instance(Instance &copyInstance);
  Instance(SequenceObject *pSeqObj, strand_modes s, int start, int length);
  Instance(string * seqid, const int &start, const int &end, 
           const strand_modes &s, string * site, string motifid, double w); // FC

  // destructor
  ~Instance(); // FC

  // inspectors
  int Start(){ return _start; };
  int Length(){ return _length;};
  strand_modes Strand(){ return _strand;};
  double Score(){ return _score; };
  string *ID(){ return _pId; };
  Sequence * Site(){ return _pSite; };
  string * PrintSite(){ return _pSiteString; };
  SequenceObject * ParentSequence(){return _pParentSequence;};
  string * PrintSeqName(){return _pSeqName;};
  // FC
  int IntStore(){return _iStore;}; // FC
  string * GetMotifID(){return _pMotifID;}; // FC
  string * InstanceID(); // (new) FC
  double GetWeight(){return _weight;}; // FC
  vector<string> * GetAllMotifs(){return _pJoinMotifIDs;}; // FC
  vector<double> * GetAllWeights(){return _pJoinWeights;}; // FC
  int NbrOcc(){return _pOcc;} // FC
  string * GetOverlaps(); // FC
  vector<Instance*>* GetOverlapInsts(){return _pOverlaps;}; // FC
  int SimilarMotifs(Instance * pInstance2); // FC
  Instance * MinStart(); // FC
  Instance * MaxEnd(); // FC


  bool CheckLeftExtension(int pos);
  bool CheckRightExtension(int pos);
  
  // adaptors
  void SetID(string *pId){ _pId = pId; return;};
  Instance * SetScore(double score){ _score = score; return this;};
  Instance * SetSeqName(string *pName){_pSeqName = pName; return this;};
  Instance * ExtendLeft(int pos);
  Instance * ExtendRight(int pos);
  // FC
  void SetOverlaps(vector<Instance*>* laps){_pOverlaps = laps; return;}; // to assign NULL...
  void SetStoreMode(int m){_iStore = m; return;}
  void NormWeight(double w){_weight /= w; return;}
  void CreateOverlaps();
  void AddInstance(Instance * pInst); // add instance/_pOcc
  void UpdateInstance(); // update other info
  void SetString(string code){_pSiteString = new string(code); return;} // assumes _pSiteString == NULL
  void _TranslateToSequence();
    
};



#endif
