#include "Cluster.h"

/******************************************************************************
  Description:  constructor
  Author_Date:  MC_2009/07/22
******************************************************************************/
Cluster::Cluster(int nbrN, double * cluster, int nbrEx1, 
				   int nbrD, double * data, int nbrEx2)
{
  double min = 2; double max = -1;
  // node membership of cluster
  _pCluster = cluster;
  _length1 = nbrN;
  _nbrNodesEx = nbrEx1;
  // compute number/min/max node membership 
  _nbrNodes = 0; 
  for(int i = 0; i < nbrN; i++)
  {
    if (cluster[i] > 0) 
    {  _nbrNodes++;
      if (cluster[i] < min) min = cluster[i];
      if (cluster[i] > max) max = cluster[i];
    }
  }
  _NodeThresh = min/max;

  min = 2; max = -1;
  // data membership of clusters
  _pClusterData = data;
  _length2 = nbrD;
  _nbrDataEx = nbrEx2; 
  // compute number/min/max membership 
  _nbrData = 0; 
  for(int i = 0; i < nbrD; i++)
  {
    if (data[i] > 0) 
    {_nbrData++;
      if (data[i] < min) min = data[i];
      if (data[i] > max) max = data[i];
    }
  }
  _DataThresh = min/max;

  _pGoodCSMotif = NULL;
  _pGoodwidthMotif = NULL;
  _pseudo = new double[4];
  for(int i = 0; i < 4; i++) _pseudo[i] = 0.001;
  _pClusterShifts = NULL;
  _pCSdistribution = NULL;
}
/******************************************************************************
  Description:  destructor
  Author_Date:  MC_2009/07/22
******************************************************************************/
Cluster::~Cluster()
{
  if(_pCluster != NULL) 
    delete[] _pCluster; 
  _pCluster = NULL;

  if(_pGoodCSMotif != NULL) 
    delete _pGoodCSMotif; 
  _pGoodCSMotif = NULL;

  if(_pGoodwidthMotif != NULL) 
    delete _pGoodwidthMotif; 
  _pGoodwidthMotif = NULL;

  if (_pseudo != NULL) 
    delete[] _pseudo;
  _pseudo = NULL;

  if(_pClusterShifts != NULL) 
    delete[] _pClusterShifts; 
  _pClusterShifts = NULL;

  if(_pClusterData != NULL)
    delete[] _pClusterData; 
  _pClusterData = NULL;

  if(_pCSdistribution != NULL)
    delete _pCSdistribution; 
  _pCSdistribution = NULL;

}


/******************************************************************************
  Description:  ReconstructMotif, given instance-codes 
                // length of instances can be fixed or variable, in both
                cases the best alignment is matched to highest prob instance,
                and PWM is extended in width for optimal CS-score
  Author_Date:  MC_2009/09/14
******************************************************************************/
void
Cluster::ReconstructMotif(InstanceSet * instanceSet)
{
  //cerr << "debug::Cluster::ReconstructMotif - begin" << endl;
  // local variable to store shift positions of each node
  _pClusterShifts = new double[_length1];
  for(int i = 0; i < _length1; i++) _pClusterShifts[i] = -1000;


  // get the (longest of the) highest probability instance(s)
  iSetIter Hiter = GetLongestHighestMemberInstance(instanceSet);

  int Hlength = (*Hiter)->Length();

  // compute the bestshift for each instance versus reference-instance
  // keep track of extreme shifts for PWM width reconstruction
  int extLeft(0), extRight(0);
  // iterate through all instances 
  iSetIter iterI = instanceSet->GetSet()->begin();
  for(int i = 0; i < _length1; i++)
  {
    if (iterI != Hiter && _pCluster[i] != 0)
    {
      // compute and store the best shift
      _pClusterShifts[i] = FUZZINCLUSIVE::BestShift(*Hiter, *iterI);
      if (_pClusterShifts[i] < extLeft) extLeft = _pClusterShifts[i];
      else if (_pClusterShifts[i] + (*iterI)->Length() - Hlength > extRight) 
             extRight = _pClusterShifts[i] + (*iterI)->Length() - Hlength;
    }
    iterI++;
  }
//cerr << "Hlength/extleft/extright = " << Hlength << "/" << extLeft << "/" << extRight << endl;

  // compute the PWM based on Hprob
  PWM * pwmMotif = BuildPWM(instanceSet, Hlength, 0);
  double H_CS = pwmMotif->GetCS();
//cerr << "Hlength : " << pwmwidth << "," << H_CS <<"/" << *pwmMotif->GetConsensus() << endl;
  delete pwmMotif; pwmMotif = NULL;

  // now search for the most optimal PWM 
  // start with the one based on Hprob, extend to right resp left 
  // keep track of the width at good CS and track acceptable further extensions
  int * pLeftExt = SearchExtensions(instanceSet, Hlength, -extLeft, -1, H_CS);
  int * pRightExt = SearchExtensions(instanceSet, Hlength, extRight, 0, H_CS);

//cerr << "check = " << pLeftExt[0] << "/" << pRightExt[0] << "&&" 
//                   << pLeftExt[1] << "/" << pRightExt[1] << endl;

  // now finally store the most optimal matrices
  // construct good-cs motif and also store the good-width motif
  _pGoodCSMotif = BuildPWM(instanceSet, 
                  Hlength + pRightExt[0] + pLeftExt[0], -pLeftExt[0]);
  _pGoodwidthMotif = NULL;
  if (pRightExt[1] > pRightExt[0] || pLeftExt[1] > pLeftExt[0])
  { _pGoodwidthMotif = BuildPWM(instanceSet, 
                   Hlength + pRightExt[1] + pLeftExt[1], -pLeftExt[1]);}
//cerr << "good-cs= : " << pwmwidth << "," << csLeft << "," << csRight << endl;
//cerr << "good-width : " << pwmwidth<< "," << widthLeft << "," << widthRight << endl;

 
  // some final reporting adjustments 
  // reset the shift versus _pGoodCSMotif instead of iterH
  if (pLeftExt[0] != 0)
  { for(int i = 0; i < _length1; i++){ _pClusterShifts[i] += pLeftExt[0];}}

  // and report the CS-distribution versus _pGoodCSMotif
//cerr << "CSdistribution: " << -extLeft << "-" << pLeftExt[0] << " && "
//<< extRight << "-" << pRightExt[0] << endl;
  //if ((-extLeft - pLeftExt[0]) != 0 || (extRight - pRightExt[0]) != 0)
  //{ BuildCSdistribution(instanceSet, Hlength + pRightExt[0] + pLeftExt[0],
  //                    -extLeft - pLeftExt[0], extRight - pRightExt[0]);}
  BuildCSdistribution(instanceSet, Hlength + pRightExt[0] + pLeftExt[0],
                      -extLeft - pLeftExt[0], extRight - pRightExt[0]);
  // cleanup
  delete[] pLeftExt; pLeftExt = NULL;
  delete[] pRightExt; pRightExt = NULL;
  //cerr << "debug::Cluster::ReconstructMotif - end" << endl;
  return;
}

/******************************************************************************
  Description:  GetLongestHighestMemberInstance(instanceSet)
  Author_Date:  MC_2009/09/16
******************************************************************************/
iSetIter 
Cluster::GetLongestHighestMemberInstance(InstanceSet * instanceSet)
{
  // get the (longest of the) highest node-membership instance(s)
  double probH = 0; int lengthH = 0; 
  int iH; iSetIter iterH;
  iSetIter iterI = instanceSet->GetSet()->begin();
  for(int i = 0; i < _length1; i++)
  { 
    if (_pCluster[i] != 0 && _pCluster[i] >= probH)
    { 
      if (_pCluster[i] > probH || 
           (_pCluster[i] == probH && (*iterI)->Length() > lengthH))
      {
        iH = i; iterH = iterI; 
        lengthH = (*iterI)->Length(); 
        probH = _pCluster[i]; 
      }
    }
    iterI++;
  }
  _pClusterShifts[iH] = 0;
  return iterH;
}

/******************************************************************************
  Description:  SearchExtensions(instanceSet, Hlength, extRight, 0);
  //int * pLeftExt = SearchExtensions(instanceSet, Hlength, -extLeft, -1, H_CS);
  //int * pRightExt = SearchExtensions(instanceSet, Hlength, extRight, 0, H_CS);
  Author_Date:  MC_2009/12/01
******************************************************************************/
int *  
Cluster::SearchExtensions(InstanceSet * instanceSet, int length, 
                          int ext, int shift, double CSref)
{
  int * pExt = new int[2]; 
  pExt[0] = 0; // extension with best CS
  pExt[1] = 0; // other interesting extension

  bool up = true;
  double CS, lastCS = CSref;

  int width = length + 1; 
  while (width <= length + ext)
  {
  //cerr << "###width = " << width  << ", extension= " << length - width << ", shift= " << shift << endl;
    PWM * pMotif = BuildPWM(instanceSet, width, shift);
    CS = pMotif->GetCS();
  //cerr << "pMotif->GetCS() = " << CS << "/" << *pMotif->GetConsensus() << endl;
    // keep track of best CS
    if ( CS > CSref) { pExt[0] = abs(length - width);}
    // or of two not-descending CS extension (to cover spacers)
    if (CS >= lastCS) { if (up) {pExt[1] = abs(length - width); } up = true;}
    else up = false;
    lastCS = CS; // added 2009/12/01
    delete pMotif; pMotif = NULL;
    width++;
    if (shift != 0) shift--; // only needed for left-extensions
  }
  return pExt;
}

/******************************************************************************
  Description:  BuildPWM : subfunction for ReconstructMotif
  Author_Date:  MC_2009/09/15
******************************************************************************/
PWM *
Cluster::BuildPWM(InstanceSet * instanceSet, int width, int HprobShift)
{
  //cerr << "debug::Cluster::BuildPWM - begin" << endl;
  Sequence * pSite = NULL;
  int nt;

  double ** pCounts = new double*[width]; 
  for (int w = 0; w < width; w++)
  { pCounts[w] = new double[4]; 
    for (int n = 0; n < 4; n++) 
      pCounts[w][n] = 0;
  } 
  // add all instances to counts in line with bestshift/HprobShift
  iSetIter iter = instanceSet->GetSet()->begin();
  for(int i = 0; i < _length1; i++)
  {
    if (_pCluster[i] != 0) 
    {
      // add the instance to counts in line with bestshift/HprobShift
      pSite = FUZZINCLUSIVE::ConvertSiteToLength(*iter, width, 
                                            _pClusterShifts[i] - HprobShift);
      for (int p = 0; p < width; p++)
      { 
        nt = (*pSite)[p];
        if (nt != -1) pCounts[p][nt] += _pCluster[i];
        else {for(int n = 0; n < 4; n++) pCounts[p][n] += _pCluster[i]/4;}
      }
      delete pSite; pSite = NULL;
    }
    iter++;
  }

  // build the pwm, also CS-score and Consensus are computed
  PWM * pMotif = new PWM(width, pCounts, _pseudo);
  // cleanup
  for (int w = 0; w < width; w++) delete[] pCounts[w]; 
  delete[] pCounts; pCounts = NULL;
  //cerr << "debug::Cluster::BuildPWM - end" << endl;

  return pMotif;
}
/******************************************************************************
  Description:  BuildCSdistribution(  ) for reporting purposes
  Author_Date:  MC_2009/09/16
******************************************************************************/
void
Cluster::BuildCSdistribution(InstanceSet * instanceSet, 
                             int length, int left, int right)
{
  list<double>* pCSdist = new list<double>;
  int width = length + 1;
  int shift = -1; double CS;
  // to the left
  while (width <= length + left + right)
  {
    PWM * pMotif = BuildPWM(instanceSet, width, shift);
    CS = pMotif->GetCS();
    pCSdist->push_front(CS);
    delete pMotif; pMotif = NULL;
    width++;
    shift--;
  }
	
  ostringstream csdist; 
  if ((int)pCSdist->size() > 0)
  {
    csdist << "(";
    list<double>::iterator iterL = pCSdist->begin();
    for(;iterL != pCSdist->end(); iterL++)
    { if (iterL != pCSdist->begin()) csdist << ",";
      csdist << setprecision(3) << (*iterL);
    }
    csdist << ")";
  }
  delete pCSdist; pCSdist = NULL;
  csdist << "[" << setprecision(3) << _pGoodCSMotif->GetCS() << "]";

  // right side
  width = length + 1;
  if (width <= length + right + left)
  { csdist << "(";
    while (width <= length + right + left)
    {
      PWM * pMotif = BuildPWM(instanceSet, width, 0);
      CS = pMotif->GetCS();
      if (width != length + 1) csdist << ",";
      csdist << setprecision(3) << CS;
      delete pMotif; pMotif = NULL;
      width++;
    }
    csdist << ")";
  }
  _pCSdistribution = new string(csdist.str()); 
  csdist.flush();
  return;
}

/******************************************************************************
  Description:  WeightedInstanceCount();
  Author_Date:  MC_2011/11/24
******************************************************************************/
int 
Cluster::WeightedInstanceCount()
{
  // sum all instance memberships and normalize with highest instance membership
  double count = 0;
  double max = -1;
  int result = -1;
  for(int i = 0; i < _length1; i++)
  { 
    if (_pCluster[i] > max) { max = _pCluster[i];} 
    count += _pCluster[i];
  }
  count /= max;
  // round to nearest integer
  if ((count - int(count)) < 0.5){result = int(count);} 
  else {result = int(count)+1;}
  return result;
}
/******************************************************************************
  Description:  WeightedmotifCount();
  Author_Date:  MC_2011/11/24
******************************************************************************/
int 
Cluster::WeightedMotifCount()
{
  // sum all motif memberships and normalize with highest motif membership
  double count = 0;
  double max = -1;
  int result = -1;
  for(int i = 0; i < _length2; i++)
  { 
    if (_pClusterData[i] > max) { max = _pClusterData[i];} 
    count += _pClusterData[i];
  }
  count /= max;
  // round to nearest integer
  if ((count - int(count)) < 0.5){result = int(count);} 
  else {result = int(count)+1;}
  return result;
}

