#include "FastaIO.h"

#include <iostream>


/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
		 
  Description:  constructor used for FuzzyClustering of MS-3.1.5 files
                the lines start with #id: or >
  Author_Date:  MC_2011/06/27
******************************************************************************/
FastaIO::FastaIO(string * fileName)
{
  // open file for reading
  _ifs.open(fileName->c_str(), ios::in);

  if (!_ifs)
  {
    // file is not open for reading
    cerr << "--Error-- FastaIO::FastaIO() unable to open file: "
      << *fileName << " for reading!" << endl;
    _isOpen = false;
    _hasNext = false;
    /* 
     * the program will not stop here
     * the user should check with IsOpen() in main program
     */
  }
  else
  {
    _isOpen = true;

    // read lines untill first line starting with "#id" or ">" is found
    while (!_ifs.eof())
    {
      getline(_ifs, _pLine, '\n');
      //cerr << "DEBUG: " << _pLine << endl;
      if (_pLine.empty() || (int)_pLine.size() < 3){ continue;}
      if ( 
           (_pLine[0] == '#' && _pLine[1] == 'i' && _pLine[2] == 'd')
		   || (_pLine[0] == '>' )) // 2011/06/27
      {  break;}
    }      
	//cerr << "DEBUG: " << _pLine << endl;
 
    if (!_ifs.eof())
    {
      _hasNext = true;
      //cerr << "DEBUG: _hasNext = true " << endl;
    }
    else
    {
      _hasNext = false;
      //cerr << "DEBUG: _hasNext = false " << endl;
    }
  }
}

/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
FastaIO::~FastaIO()
{
  // close file
  if (_ifs.good())
    _ifs.close();

  return;
}


/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
void
FastaIO::Close()
{
  _ifs.close();
  _isOpen = false;
  return;
}


/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:  load a sequence from FASTA file
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  Date:         MC_2011/07/03 allow phylofasta file as input in MS.
                // mark this function is not used in 5.0.5, 
                only changed here to keep code the same with 3.1.5  
******************************************************************************/
SequenceObject *
FastaIO::NextSequence()
{

  if (_isOpen && _hasNext)
  {

    // create empty string to store sequence
    string *pSeqStr = new string("");

    string *pDescStr = new string(_pLine);
    
    // cerr << "DEBUG: " << *pDescStr;

    while (!_ifs.eof())
    {
      // read next line
      getline(_ifs, _pLine, '\n');

      // move to next line if line is empty and skip >>gene-group lines 
      if (_pLine.empty() || (_pLine[0] == '>' && _pLine[1] == '>'))
        continue;

      if (_pLine[0] == '>')
      {
        break;
      }
      else
      {
        string::size_type pos = 0;
        string tokens = " \t\n\r\f";
        while ((pos = _pLine.find_first_of(tokens)) != string::npos)
        {
          _pLine.erase(pos, 1);
        }
        // append line to seqStr;
        pSeqStr->append(_pLine);

      }
    }

    if (!_ifs.eof())
    {
      _hasNext = true;
    }
    else
    {
      _hasNext = false;
    }

    // create new sequence object
    // cerr << *seqStr << endl;

    SequenceObject *pSeqObj = new SequenceObject(pSeqStr);
    pSeqObj->SetDescription(pDescStr);

    delete pSeqStr;
    delete pDescStr;

    return pSeqObj;

  }
  else
  {
    // sequence was not open anymore 
    return NULL;
  }
}

/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
bool
FastaIO::IsOpen()
{
  return _isOpen;
}

/******************************************************************************
  Method: 
  Class:        
  Arguments: 
  
  Description:
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
bool
FastaIO::HasNext()
{
  return _hasNext;
}

/******************************************************************************
  Description:  ReadLine : returns current _pLine  
                (line may but does not necessarily start with >)
				and set cursor on next non-empty line 
  Author_Date:  MC_2009/12/07
******************************************************************************/
string
FastaIO::ReadLine()
{
  //cerr << "debug - FastaIO::ReadLine - begin" << endl;
  string thisline = _pLine;
	
  _hasNext = false;
  // proceed untill next valid input line
  while (!_ifs.eof())
  {
    getline(_ifs, _pLine, '\n');
    if (_pLine.empty())
      continue;
    if ( _pLine[0] == '>' || 
        (_pLine[0] == '#' && _pLine[1] == 'i' && _pLine[2] == 'd')
       ) _hasNext = true;
    break; // other not-empty line not starting with > or #id: 
  }
  if (_ifs.eof()) _isOpen = false;
  //cerr << "debug - FastaIO::ReadLine - end" << endl;
  return thisline;
}
/******************************************************************************
  Description:  CheckLineStart()
  Author_Date:  MC_2011/04/11
******************************************************************************/
bool 
FastaIO::CheckLineStart(string Start)
{
  string::size_type pos = 0;
  string tokens = " \t\n\r\f";
  if (_pLine.empty() || ((pos = _pLine.find_first_of(tokens)) == 0))
  {
    cerr << "Skip empty line." << endl;
    // set cursor ready on next non-empty line
    ReadLine(); 
	return 1;
  }
  if (_pLine[0] == Start[0]) // e.g. lines starting with #
  {
    cerr << "Skip line " << _pLine << endl;
    // set cursor ready on next line
    ReadLine(); 
	return 1;
  } 
  return 0;
}
/******************************************************************************
  Description:  ReadInstance - different formats allowed
  Author_Date:  MC_2009/07/22
******************************************************************************/
Instance *
FastaIO::ReadInstance(string motifid)
{
  //cerr << "debug - FastaIO::ReadInstance - begin" << endl;
  Instance * pInstance = NULL;
  char * pcSeqid = new char[256];
  int iStart, iEnd;
  double dWeight = -1;
  char * pcStrand = new char[2];
  char * pcSite = new char[256];
  string * Seqid(NULL), * Site(NULL);

  if (
      /*MotifSampler-old*/
//yneN	MotifSampler	misc_feature	241	259	15535.2	+	.	id "box_1_1_wCwTCkwTnnAATmnnTnA"; site "ACATATTTGAAATGCATGC";
      sscanf(_pLine.c_str(), "%s %*s %*s %d %d %*f %s %*s %*s %*s %*s %s", 
                     pcSeqid,&iStart,&iEnd, pcStrand, pcSite) == 5 ||
      /*MotifSampler*/ 
//yneN	MotifSampler	misc_feature	241	259	15535.2	+	id "box_1_1_wCwTCkwTnnAATmnnTnA"; site "ACATATTTGAAATGCATGC";
      sscanf(_pLine.c_str(), "%s %*s %*s %d %d %*f %s %*s %*s %*s %s", 
                      pcSeqid, &iStart, &iEnd, pcStrand, pcSite) == 5  ||
      /*PhyloMotifSampler*/ 
      sscanf(_pLine.c_str(), "%s %d %d %*f %s %*s %s", pcSeqid, &iStart, 
                                &iEnd, pcStrand, pcSite) == 5 ||
     /*minimal format with weight*/ 
//yneN	241	259	+	"ACATATTTGAAATGCATGC"	0.95
      sscanf(_pLine.c_str(), "%s %d %d %s %s %lf", pcSeqid, &iStart, 
                         &iEnd, pcStrand, pcSite, &dWeight) == 6 ||
     /*minimal format*/ 
//yneN	241	259	+	"ACATATTTGAAATGCATGC"
      sscanf(_pLine.c_str(), "%s %d %d %s %s", pcSeqid, &iStart, 
                                &iEnd, pcStrand, pcSite) == 5 )
  {
    Seqid = new string(pcSeqid);
    Site = new string(pcSite);
    string::size_type pos = 0;
    while ( (pos = Site->find_first_not_of("acgtnACGTN")) != string::npos )
      Site->erase(pos,1); // erase non-code char (such as " or ;)
    if ( iStart >= 0 && 
         (iEnd - iStart +1) == (int)Site->size() &&
         ( pcStrand[0] == '+' || pcStrand[0] == '-'))
    {
      if (dWeight == -1) dWeight = 1; // set uniform weight
      if (dWeight > 0)
      {
        // create new instance 
        if (pcStrand[0] == '+') 
          pInstance = new Instance(Seqid, iStart, iEnd, plus_strand, Site, 
								   motifid, dWeight);
        else 
          pInstance = new Instance(Seqid, iStart, iEnd, minus_strand, Site,
								   motifid, dWeight);
/*  cerr << "debug: instance: ";
        if (pInstance != NULL)
        {
		  // print id for debug
		  string * id =pInstance->InstanceID();
		  cerr << "instanceid = " << *id << endl;
		  delete id; id = NULL;
	    }
		  cerr << "debug-end" << endl;
*/

      }
    }
    else 
    {
     cerr << "--The Start, Length or Strand description is not correct in " << _pLine << endl;
    }
  }
  // cleanup
  if (Seqid != NULL) delete Seqid; Seqid = NULL;
  if (Site != NULL) delete Site; Site = NULL;
  delete[] pcSeqid; 
  delete[] pcStrand;
  delete[] pcSite;
	
 if (pInstance == NULL)// throw error handled in MotifClusterRun.cpp
 { cerr << "--Error-FastaIO::ReadInstance() : error in " << _pLine << endl;}

  // set cursor ready on next not-empty line
  ReadLine(); 
  //cerr << "debug - FastaIO::ReadInstance - end" << endl;
  return pInstance;
}
